// This progamme uses a set of band pins or ICOM CIV
// generate control signals for relays, amplifiers etc. Uses CI-V 9600 baud control
// The frequancy band and status is shown on an I2C LCD
// Based on code from Nick Gammon, ON7EQ, DM2RM, VE1ZAC and OK1CDJ

#include <avr/pgmspace.h>
#include "definitions.h"              // File with all the defines in it 
#include <SoftwareSerial.h>           // Software serial code
#include <EEPROM.h>                   // EEPROM to memorize antenna / band config
#include <Wire.h>                     // Wire is required for i2c LCD Display
#include <NoiascaLiquidCrystal.h>     // download from https://werner.rothschopf.net/202009_arduino_liquid_crystal_intro.htm 
#include <NoiascaHW/lcd_PCF8574.h>    // I2C PCF Adapter - works with software serial

// Globals
const char* mode[] = {"LSB", "USB", "AM ", "CW ", "FSK", "FM ", "WFM"};
String modes;
const uint32_t decMulti[] = {1000000000, 100000000, 10000000, 1000000, 100000, 10000, 1000, 100, 10, 1};
uint32_t readtimeout;                 // Serial port read timeout
uint8_t read_buffer[12];              // Read buffer
uint32_t frequency;                   // Current frequency in Hz
uint32_t timer;                       // timeout timer 
uint8_t radio_address;                // Transiever address
int unsigned memotime = (1000);       // delay (ms) before selected antenna is memorized
uint8_t DataMode;                     // The input mode can be CIV or Band data lines
 
bool buttonpressed = false;           // true if any button pressed
bool leftpressed = false;             // true if left button pressed
bool rightpressed = false;            // true if right button pressed
uint8_t numAntennas;                  // how many antenna relays in our system, 2-8
uint16_t bandADC[16];                 // Array of adc thresholds to convert to band number

int incoming;
uint32_t DisplayTime = 0;             // timer display refresh
uint32_t ButtonTime = 0;              // Timer button depressed
byte memorize = (0);                  // if = 1, them position to be memorzied
int8_t menuIdx = -1;
int8_t menuMaxIdx = 3;
String menuText[] = {"Band Detection", "CIV Address", "Exit Menu", "Reset All"};

// Set up the software serial port
SoftwareSerial CIV(RX_PIN, TX_PIN); 

// Initialise LCD
LiquidCrystal_PCF8574_base lcd(LCD_ADDR, LCD_COL, LCD_ROW); 
uint8_t numBands;                     // Number of bands  
int8_t currentBand = 0;               // Currently selected band
int8_t lastBand = 255;                // Previously selected band ensure invalid for first run

/*-------------------------------------------------------------------------*/
void setup() {
  
Serial.begin(9600); // Used by Bluetooth module
Serial.setTimeout(TIMEOUT);

  lcd.init();                         // LCD display init
  lcd.backlight();                    // LCD backlight
  lcd.begin(16, 2);                   // LCD Rows and Columns
  lcd.clear();

  // Set up relay output pins
  // How many bands are there?
  numAntennas = sizeof(RelayPin) / sizeof(uint8_t);

  for(uint8_t i = 0; i < numAntennas; i++) 
    {
    pinMode(RelayPin[i], OUTPUT);  // Set pin to an output
    digitalWrite(RelayPin[i], 0);  // Ensure a zero initial state       
    }  

  // Set up buttons
  pinMode(LeftButtonPin, INPUT_PULLUP);
  pinMode(RightButtonPin, INPUT_PULLUP);
  digitalWrite(LeftButtonPin, 1);
  digitalWrite(RightButtonPin, 1);

  // Set up band select input pins
  pinMode(Band0Pin, INPUT);    
  pinMode(Band1Pin, INPUT);
  pinMode(Band2Pin, INPUT);
  pinMode(Band3Pin, INPUT);
   
  // Print a message to the LCD.
  lcd.setCursor(0, 0);
  lcd.print(F("Antenna Swtich 8"));
  lcd.setCursor(0, 1);
  lcd.print(F("CIV/Data/Voltage"));

  delay(1000);
  lcd.clear();
  lcd.print(F("V1 - G0MJW 2023"));  
  delay(1000);
  lcd.clear();

  // How many bands are there?
  numBands = sizeof(FrequencyBand) / sizeof(bandDefs_t);

  // Set up the ADC array for band detection by voltage
  setVoltsToBand();

  // First run check
  if(EEPROM.read(0) != 1 && EEPROM.read(1) != 2 && EEPROM.read(2) != 3 && EEPROM.read(3) != 4)
    {
    lcd.println(F("First Run"));
    lcd.println(F("Initialising"));  
    EEPROMinit();
    delay(1000);
    lcd.clear();     
    }   

  // Start port running at 9k6
  CIV.begin(9600);
  CIV.setTimeout(TIMEOUT);

  // Find our radio - variable so we can change it if we want
  radio_address = EEPROM.read(EEPROM_RADIO_ADDR);

  // Band detection mode 
  DataMode = EEPROM.read(EEPROM_MODE_FLAG);
 
  // Enter setup if both left & right buttons pressed at startup
  if (!digitalRead(LeftButtonPin) && !digitalRead(RightButtonPin))
    initialMenu();
 
}
/*--------------------------------------------------------------------------*/
void loop() 
{
  static const unsigned long REFRESH_INTERVAL = 1000; // ms
  static unsigned long lastRefreshTime; // LCD refrech timer in loop

 #ifdef DEBUG
 //  uint32_t timestamp = millis();    
 #endif  

   ReadButtons();
   // Update LCD once a second and check band
   if(millis() - lastRefreshTime >= REFRESH_INTERVAL)
    {      
    lastRefreshTime += REFRESH_INTERVAL;   
    // Check for band change - CIV or Band depending
    if((DataMode == CIV_MODE) || (DataMode == BT_MODE))
      {
      if(CIVbandchange())
        changeBand(); 
      }
     else
      {
      if(DATAbandchange())
        changeBand(); 
       }
          
    // Update the LCD
    updateLCD();       
    }                         
  }
/*-------------------------------------------------------------------------*/
bool DATAbandchange() {
  // Checks for band changes periodically from the IO lines
  // returns true if band changed 

  // Here we need to decode the frequency to one of the bands
  // Find which band by running through the band list
  // First matching band taken
  // const long FreqencyBand[numBands][3]       // Freq kHz from, to, band number
  uint8_t Band = 0;
  
  // Read the data lines or band volatge
  if(DataMode == BAND_MODE)
    Band = (digitalRead(Band3Pin) << 3)
          + (digitalRead(Band2Pin) << 2) 
          + (digitalRead(Band1Pin) << 1) 
          + digitalRead(Band0Pin);
  else
    Band = VoltsToBand(analogRead(VoltInPin)); // Using the ADC

  currentBand = 0; // Set to an invalid band to start
  for(uint8_t i = 0; i < numBands; i++)      
    if(Band == FrequencyBand[i].bandData)
      {
      currentBand = i; 
      break;
      }
          
  // Check for band change
  if(currentBand != lastBand) {    
    lastBand = currentBand;
    return true; 
    }     
  else  
    return false;  
}
/*-------------------------------------------------------------------------*/
bool CIVbandchange() {
  // Checks for band changes periodically sending command to read main vfo band
  // returns true if band changed 
  
  sendCivRequest(CMD_READ_FREQ);  
  processCivMessages();
  delay(50); // some radios are a little slow and need a delay
  sendCivRequest(CMD_READ_MODE);
  processCivMessages();

  // Here we need to decode the frequency to one of the bands
  // Find which band by running through the band list
  // First matching band taken
  // const long FreqencyBand[numBands][3]       // Freq kHz from, to, band number

  currentBand = numBands + 1; // Set to an invalid band if frequency is out of range of filters 
  for(uint8_t i = 0; i < numBands; i++)      
    if(frequency >= FrequencyBand[i].startFreq && frequency <= FrequencyBand[i].stopFreq)
      {
      currentBand = i; 
      break;
      }
          
  // Check for band change
  if(currentBand != lastBand) {    
    lastBand = currentBand;
    return true; 
    }     
  else  
    return false;  
}
/*--------------------------------------------------------------------------*/
void updateLCD() {
  // Update LCD
  char text[21];  

  // Frequency
  lcd.setCursor(0, 0);  
  uint16_t MHz = frequency / 1000000L;            // Convert to MHz  
  uint16_t kHz = (frequency % 1000000L) / 1000;   // The decimal part in kHz  
  uint16_t Hz = (frequency % 1000000L) % 1000;    // The decimal part in Hz  

  if(DataMode == CIV_MODE || DataMode == BT_MODE)   
    sprintf(text,"%3u.%03u,%03u MHz ",MHz, kHz, Hz);
  else if(DataMode == BAND_MODE) 
    sprintf(text,"B0-B3 lines %02u  ",FrequencyBand[currentBand].bandData);
  else if(DataMode == VOLT_MODE)
    sprintf(text,"Voltage band %02u ",FrequencyBand[currentBand].bandData);
    
  lcd.print(text); 

  // Band and mode and IO
  lcd.setCursor(0, 1);   
  strncpy(text, FrequencyBand[currentBand].bandName, 4);
  text[4] = NULL;
  lcd.print(text);   
  lcd.print(" ");   
  lcd.print(modes);  
  lcd.print(F("  Ant "));   
  lcd.print(ReadEEPROMvalues(currentBand)); 
 } 
 /*-------------------------------------------------------------------------*/
void changeBand() 
{
  // Set the relay as pointed to in the EEPROM array 
  uint8_t Antenna = ReadEEPROMvalues(currentBand);
  
  // Validity check
  if(Antenna >=0 && Antenna <= numAntennas)
    SelectAntenna(Antenna);   
}
/*-------------------------------------------------------------------------*/ 
void sendCivRequest(uint8_t requestCode)
{
  uint8_t req[] = {START_BYTE, START_BYTE, radio_address, CONTROLLER_ADDRESS, requestCode, STOP_BYTE};

  for (uint8_t i = 0; i < sizeof(req); i++) 
    {
    if(DataMode == CIV_MODE)
      CIV.write(req[i]);     
    if(DataMode == BT_MODE) 
      Serial.write(req[i]);  
     }
}
/*-------------------------------------------------------------------------*/ 
uint8_t readLine(void)
{
  uint8_t byte;
  uint8_t counter = 0;
  uint32_t ed = readtimeout;

  while (true)
  {
    if(DataMode == CIV_MODE)
      while (!CIV.available()) 
      {
       if (--ed == 0)
        return 0;
      }
    if(DataMode == BT_MODE)
     while (!Serial.available()) 
      {
      if (--ed == 0)
        return 0;
      }
    
    ed = readtimeout;
    // If data available read a byte
    if(DataMode == CIV_MODE)
      byte = CIV.read();
    if(DataMode == BT_MODE) 
      byte = Serial.read(); 
      
    if (byte == 0xFF)continue; //TO DO skip to start byte instead

    read_buffer[counter++] = byte;
    if (STOP_BYTE == byte) break;

    if (counter >= sizeof(read_buffer))return 0;
  }
  return counter;
}
/*-------------------------------------------------------------------------*/ 
void findFrequency(void)
{  
  frequency = 0;
  //FE FE E0 42 03 <00 00 58 45 01> FD ic-820
  //FE FE 00 40 00 <00 60 06 14> FD ic-732
  for (uint8_t i = 0; i < 5; i++) {
    if (read_buffer[9 - i] == 0xFD)continue; //spike

    frequency += (read_buffer[9 - i] >> 4) * decMulti[i * 2];
    frequency += (read_buffer[9 - i] & 0x0F) * decMulti[i * 2 + 1];
  }
}
/*-------------------------------------------------------------------------*/ 
void processCivMessages()
{

  while ( (CIV.available() && (DataMode == CIV_MODE)) || (Serial.available() && (DataMode == BT_MODE))) 
  //while ( CIV.available() ) 
    {
    uint8_t knowncommand = 1;
    uint8_t r;
    if (readLine() > 0) {
      if (read_buffer[0] == START_BYTE && read_buffer[1] == START_BYTE) {
        if (read_buffer[3] == radio_address) {
          if (read_buffer[2] == BROADCAST_ADDRESS) {
            switch (read_buffer[4]) {
              case CMD_TRANS_FREQ:
                findFrequency();
                break;
              case CMD_TRANS_MODE:
                findMode();
                break;
              default:
                knowncommand = false;
            }
          } else if (read_buffer[2] == CONTROLLER_ADDRESS) {
            switch (read_buffer[4]) {
              case CMD_READ_FREQ:
                findFrequency();
                break;
              case CMD_READ_MODE:
                findMode();
                break;
              default:
                knowncommand = false;
            }
          }
        } else {}
      }
    }
  }  
}
/*-------------------------------------------------------------------------*/ 
void radioSetMode(uint8_t modeid, uint8_t modewidth)
{
  uint8_t req[] = {START_BYTE, START_BYTE, radio_address, CONTROLLER_ADDRESS, CMD_WRITE_MODE, modeid, modewidth, STOP_BYTE};

  for (uint8_t i = 0; i < sizeof(req); i++) {
    if(DataMode == CIV_MODE)
      CIV.write(req[i]);
    if(DataMode == BT_MODE)
      Serial.write(req[i]);   
  }
}
/*-------------------------------------------------------------------------*/ 
void findMode(void)
{
  // Return string should be like FE FE E0 42 04 <00 01> FD
  modes = mode[read_buffer[5]];
  //read_buffer[6] -> 01 - Wide, 02 - Medium, 03 - Narrow
}
/*-------------------------------------------------------------------------*/
void SelectAntenna(uint8_t Antenna) 
{ 
  // Process selected antenna 
  // Note an antenna value of 0 or > 8 implies no antenna selected
  // As the array for the relays is zero based, but antennas are
  // numbered from 1 to 8, we need to add 1 to the test

  for(int i = 0; i < numAntennas; i++)  
    digitalWrite(RelayPin[i], (Antenna == (i+1))); // +1 because our antennas are 1 to 8  
}

/*-------------------------------------------------------------------------*/
uint8_t ReadEEPROMvalues(int Addr) 
{ 
  // Read Antenna stored in EEPROM.   
  if(Addr < 0 || Addr > numBands) 
    return 0;
  else
    return( EEPROM.read(Addr + EEPROM_OFFSET));  
}

/*-------------------------------------------------------------------------*/
void WriteEEPROMvalues(int Addr, uint8_t Antenna) 
{ 
  // Write Antenna to EEPROM. 
    if(Addr < 0 || Addr > numBands) 
      return;
    else
      EEPROM.write(Addr + EEPROM_OFFSET, Antenna);  
}

/*-------------------------------------------------------------------------*/
void initialMenu() 
{ 
  // Hold both buttons for 3 seconds to reset or release
  // to enter menu setup mode   

  uint32_t lastTime; // For timer in menu loops
  bool menu_mode = true;

  lcd.clear(); 
  lcd.print(F("HOLD 3s for Menu"));
  lcd.setCursor(0, 1);
  lcd.print(F("Release to exit "));

  // Check for button held or released
  lastTime = millis(); 
  while(millis() - lastTime < 3000L)
    {       
    // Check for button release
    if (digitalRead(LeftButtonPin) || digitalRead(RightButtonPin))
      {
      menu_mode = false;      
      break;
      }
    }
  lcd.clear();  
  if(menu_mode)
    menuMode();       
}
/*-------------------------------------------------------------------------*/
void ReadButtons() 
  { 
  // Detect antenna select buttons depressed

  int8_t Antenna;
    
  if (!digitalRead(LeftButtonPin) && !buttonpressed)
    { // Left        
    ButtonTime = millis();
    buttonpressed = true;     
    memorize = 1;  
    Antenna = ReadEEPROMvalues(currentBand) - 1;  // Current value decremented
    if (Antenna < 0)
      Antenna = numAntennas;
    SelectAntenna(Antenna);
    }

  if (!digitalRead(RightButtonPin) && !buttonpressed)
    { // Right       
    ButtonTime = millis();
    buttonpressed = true;
    memorize = 1; 
    Antenna = ReadEEPROMvalues(currentBand) + 1; // Current value incremented
    if (Antenna > numAntennas)
      Antenna = 0;
    SelectAntenna(Antenna);
    }  

  // Timers check
  if (millis() - ButtonTime > 300L)
   // Button press debounce
    buttonpressed = false;  

  // Memorize antenna after delay
  if ((millis() - ButtonTime > memotime) && (memorize == 1))
    { 
    memorize = 0;
    lcd.clear();
    lcd.print(F("Memorising ")); 
    lcd.print(Antenna); 
    WriteEEPROMvalues(currentBand, Antenna);
    delay(1000);
    lcd.clear();
  }
}
/*-------------------------------------------------------------------------*/
void EEPROMinit() 
{
  // Resets to default
  for(uint8_t i = 0; i < numBands; i++)   
      {
      int addr = EEPROM_OFFSET + i;
      uint8_t val = FrequencyBand[i].defaultAntenna;
      EEPROM.update(addr, val);
      }     
     // CIV Address
     radio_address = RADIO_ADDRESS;
     EEPROM.update(EEPROM_RADIO_ADDR, RADIO_ADDRESS);

     // Band detection mode - default
     DataMode = BAND_MODE;
     EEPROM.update(EEPROM_MODE_FLAG, BAND_MODE);    

     // Key to check first time programme, 1,2,3,4
     EEPROM.update(0, 1); 
     EEPROM.update(1, 2);
     EEPROM.update(2, 3); 
     EEPROM.update(3, 4);          
}
/*-------------------------------------------------------------------------*/
void menuMode() 
{
 // Setup menu
 unsigned long timeOut;
 bool timeout = true;
 bool exitmenu = false;

 // First wait for buttons to be released, or exit
 lcd.clear();
 lcd.print(F("Release buttons ")); 
 timeOut = millis() + 10000L;
 while(millis() < timeOut) 
  {
  if (digitalRead(LeftButtonPin) &&  digitalRead(RightButtonPin))
    {
    timeout = false;
    break;  
    }
  }
  lcd.clear();
  if(timeout)
    return;
 
 // If here we are ready for menu items loop 
 lcd.print(F("Select Menu"));
 while(!exitmenu)
   {
   if(getButton(LeftButtonPin)) 
     { 
     menuIdx = 0;
     leftpressed = true;
     }

   if(menuIdx > -1) 
     {
     // show the current menu title    
     displayMenu();
   
     while(!exitmenu) 
       { 
       if(getButton(LeftButtonPin) && !leftpressed) 
         { // if button is still pressed do nothing
           menuIdx += 1;
           if(menuIdx > menuMaxIdx) 
             menuIdx = 0; // reset to first menu item
           displayMenu();           
           leftpressed = true;
         } 
       else if(!getButton(LeftButtonPin) && leftpressed) 
         leftpressed = false;
            
         if(getButton(RightButtonPin)) 
           {
           exitmenu = execMenu(); 
           menuIdx = -1;  
           delay(1000);
           lcd.clear();
           lcd.print(F("Select Menu"));         
           }
        }
      menuIdx = -1;
      }
      // do some other code here
   }    
} 
/*-------------------------------------------------------------------------*/
bool getButton(int Btn) 
{
  // Checks if button pressed
  bool ret = false;
  if( digitalRead(Btn) == LOW ) 
    {
    delay(100);
    if( digitalRead(Btn) == LOW )  
      ret = true;        
    }
    return ret;
}
/*-------------------------------------------------------------------------*/
bool execMenu() 
  {
  unsigned long timeOut;
  
  if(menuIdx == 0)    // Change between CIV and BAND modes
    {
    // Toggle modes
    lcd.setCursor(0,1); 
    
    if(DataMode == CIV_MODE)
      {
      DataMode = BAND_MODE;
      lcd.print(F("Set Binary Mode "));
      
      // Set up band select input pins
      pinMode(Band0Pin, INPUT);    
      pinMode(Band1Pin, INPUT);
      pinMode(Band2Pin, INPUT);
      pinMode(Band3Pin, INPUT);      
      }
    else if(DataMode == BAND_MODE)
      {
      DataMode = VOLT_MODE;
      lcd.print(F("Set Volt Mode   "));
      }
    else if(DataMode == VOLT_MODE)
      {
      DataMode = BT_MODE;
      lcd.print(F("Set Blutooth CIV"));
      }
    else if(DataMode == BT_MODE)    
      {
      DataMode = CIV_MODE;
      lcd.print(F("Set CIV Mode    "));
      }          
      
    // Save to EEPROM
    EEPROM.update(EEPROM_MODE_FLAG, DataMode);
    }

  if(menuIdx == 1)    // Set CIV address
    {
    char buf[3]; 
    lcd.clear();
    lcd.print(F("Left to increment"));       
    lcd.print(F("Right to store"));
    delay(1000);
    lcd.clear();
    lcd.print(F("CIV = "));
    lcd.setCursor(7,0);
    sprintf (buf, "%02x", radio_address);  
    lcd.print(buf);       

    timeOut = millis() + 5000L;
      while(millis() < timeOut)
      {
      if(getButton(LeftButtonPin)) 
        {
        timeOut = millis() + 5000L; // Reset timeout
        if(radio_address < 255)
          radio_address += 1;
        else
          radio_address = 0;                
        lcd.setCursor(7,0);
        sprintf (buf, "%02x", radio_address);  
        lcd.print(buf);       
        }
      if(getButton(RightButtonPin)) // Store
          {
          EEPROM.update(EEPROM_RADIO_ADDR, radio_address);
          lcd.clear();
          lcd.print(radio_address,HEX);
          lcd.print(F(" Stored"));
          break; 
          }          
        }       
      }    

  if(menuIdx == 2)    // Escape menus
    {
    lcd.print(F("Exiting"));
    return true; // just exits
    }

  if(menuIdx == 3)    // RESET memory in EEPROM   
    {
    lcd.print(F("  RESETTING ... "));    
    EEPROMinit();     
    }     
  return false;
  }
/*-------------------------------------------------------------------------*/
void displayMenu()
{
     lcd.clear();
     lcd.print(menuText[menuIdx]);
     lcd.setCursor(0,1);
     if(menuIdx == 0)
      {
      if(DataMode == CIV_MODE)    
        lcd.print(F("Current = CIV ")); 
      else if(DataMode == BAND_MODE)
        lcd.print(F("Current = Band"));  
      else if(DataMode == VOLT_MODE)
        lcd.print(F("Current = Volt")); 
      else if(DataMode == BT_MODE)
        lcd.print(F("Current = BLTH")); 
      }
      if(menuIdx == 1) 
        {
          lcd.print(F("Current = "));
          lcd.print(radio_address,HEX);        
        }     
}
/*-------------------------------------------------------------------------*/
 uint8_t VoltsToBand(uint16_t adc) { 
  // Decodes from band voltage measurement to band
  for(int i = 0; i < 16; i++)
    if(adc <  bandADC[i])
      return (i);
    
  // Should never get here
  return (0);          
  }
/*-------------------------------------------------------------------------*/
void TestAntenna() 
{ 
  // Exercises the antenna outputs 
  for(int i = 0; i < numAntennas; i++)
    {   
    digitalWrite(RelayPin[i], 1);
    delay(500);
    digitalWrite(RelayPin[i], 0);
    }
}
/*-------------------------------------------------------------------------*/
void setVoltsToBand() 
    { 
    // This takes the voltage calibration from the header file
    // and sets ADC thresholds for band selection. It takes the 
    // mid point between each band as this threshold 
    // bandVoltageDefs[], ADC_REF, and ADC_SCALE are defined in the header

    float adc_conv;
    uint16_t adc_value;

    // The ADC to voltage conversion
    adc_conv = 1024.0 / (ADC_REF * ADC_SCALE);    
    
    for(int i = 0; i < 15; i++)
      {
      // Calculate the high thresholds for the band volage adc threshold 
      // Then we can step through and once the ADC reading is below it, 
      // we have our band
      adc_value = int(bandVoltageDefs[i] * adc_conv);  
      if(adc_value > 1024) // Cap it
        adc_value = 1024; 
      bandADC[i] = adc_value;   
      }     
  }
/*-------------------------------------------------------------------------*/
