// Header file for Arduino antenna switch

// Pins D2 and D3 are used to talk to the CIV line
// Pins D8 to D13 are the relay controls
// Pins A0, A1, A2, A3 and A7 are for the Band signals
// Pins A4 and A5 are I2C for the LCD display
// D4 and DD are buttons.

#define BAND_OUTPUT // Uses the band pins as output rather than inputs - Code TBD         

#define EEPROM_OFFSET         8       // first address in EEPROM to use for band data
#define EEPROM_RADIO_ADDR     4       // Address for CIV address
#define EEPROM_MODE_FLAG      5       // Address for mode (band/CIV) flag

#define TX_PIN                3       // Pin for CIV serial data
#define RX_PIN                2       // Pin for CIV serial data

// LCD Parameters
#define LCD_ADDR            0x27      // I2C LCD address, 3F, 27...
#define LCD_COL               16      // LCD columns, typically 16 or 20
#define LCD_ROW                2      // LCD rows, typically 2 or 4

// IO pin definitions
#define Band0Pin              A0      // Pin for band bit 0
#define Band1Pin              A1      // Pin for band bit 1
#define Band2Pin              A2      // Pin for band bit 2
#define Band3Pin              A3      // Pin for band bit 3
#define VoltInPin             A7      // Pin for band voltage input
#define LeftButtonPin          4      // Pin for left button
#define RightButtonPin         5      // Pin for right button

// Array to hold relay pin assignments 
// These are on D6 to D13 - D13 first for better layout
const uint8_t RelayPin[] = {13, 6, 7, 8, 9, 10, 11, 12};  

// Mode Setting - Default operating mode, 0 for BAND, 1 for CIV, 2 for Voltage, 3 for CIV over Bluetooth module
#define BAND_MODE               0     // Band lines
#define CIV_MODE                1     // CIV software serial
#define VOLT_MODE               2     // Yaesu Voltage level
#define BT_MODE                 3     // Same as CIV mode but with Bluetooth module

#define ADC_REF                 5.0   // ADC reference voltage
#define ADC_SCALE               1.0   // Scaling factor if there is a resistive dividor for voltage mode

// CIV settings
#define RADIO_ADDRESS        0x94     // Default - IC7300
#define CIV_BAUDRATE         9600     // CIV baud rate
#define TIMEOUT                20     // Timeout for CIV connections
#define BROADCAST_ADDRESS    0x00     // Broadcast address
#define CONTROLLER_ADDRESS   0xE0     // Controller address
#define START_BYTE           0xFE     // Start byte
#define STOP_BYTE            0xFD     // Stop byte
#define CMD_TRANS_FREQ       0x00     // Transfers operating frequency data
#define CMD_TRANS_MODE       0x01     // Transfers operating mode data
#define CMD_READ_FREQ        0x03     // Read operating frequency data
#define CMD_READ_MODE        0x04     // Read operating mode data
#define CMD_WRITE_FREQ       0x05     // Write operating frequency data
#define CMD_WRITE_MODE       0x06     // Write operating mode data
#define MODE_TYPE_LSB        0x00     // Mode data for LSB
#define MODE_TYPE_USB        0x01     // Mode data for USB
#define MODE_TYPE_AM         0x02     // Mode data for AM
#define MODE_TYPE_CW         0x03     // Mode data for CW
#define MODE_TYPE_RTTY       0x04     // Mode data for RTTY
#define MODE_TYPE_FM         0x05     // Mode data for FM

// Band definitions in struct
struct bandDefs_t
  {
  uint32_t startFreq;
  uint32_t stopFreq;
  uint8_t  bandData;
  uint8_t defaultAntenna;
  char  bandName[4];
 };

// Amateur band definitions
const bandDefs_t FrequencyBand[] = {
   {          0,          0,  0,  0, "None"},  //  No frequency data
   {    1810000,    2000000,  1,  1, "160m"},  //  160m
   {    3500000,    4000000,  2,  1, " 80m"},  //   80m  
   {    7000000,    7300000,  3,  1, " 40m"},  //   40m
   {    5298000,    5403000,  3,  1, " 60m"},  //   60m
   {   10100000,   10150000,  4,  1, " 30m"},  //   30m
   {   14000000,   14350000,  5,  2, " 20m"},  //   20m
   {   18068000,   18168000,  6,  2, " 17m"},  //   17m
   {   21000000,   21450000,  7,  2, " 15m"},  //   15m
   {   24890000,   24990000,  8,  2, " 12m"},  //   12m
   {   28000000,   29700000,  9,  3, " 10m"},  //   10m
   {   50000000,   54000000, 10,  4, "  6m"},  //    6m
   {   70000000,   72000000, 11,  4, "  4m"},  //    4m
   {  144000000,  148000000, 12,  5, "  2m"},  //    2m
   {  222000000,  225000000,  0,  5, "1.2m"},  // 1.25m
   {  420000000,  450000000, 13,  6, "70cm"},  //  70cm
   {  902000000,  928000000,  0,  7, "33cm"},  //  33cm
   { 1240000000, 1325000000, 14,  7, "23cm"},  //  23cm
   { 2300000000, 2450000000, 15,  8, "13cm"},  //  13cm
   { 3400000000, 1350000000, 15,  8, " 9cm"},  //   9cm
   { 5650000000, 5850000000, 15,  8, " 6cm"},  //   6cm
   {10000000000, 1047500000, 15,  8, " 3cm"},  //   3cm
  
  // Sweep up for non-amateur bands
   {          0,      30000,  0,  1, " VLF"},   //  VLF  
   {     148500,     283500,  0,  1, "  LW"},   //  LW
   {      30000,     300000,  0,  1, "  LF"},   //  LF
   {     530000,    1710000,  0,  1, "  MW"},   //  MW
   {     300000,    3000000,  0,  1, "  MF"},   //  MF
   {    3000000,   30000000,  0,  4, "  HF"},   //  HF
   {   76000000,  108000000,  0,  4, " BFM"},   //  FM Broadcast
   {  108000000,  137000000,  0,  4, " AIR"},   //  Airband
   {   30000000,  300000000,  0,  5, " VHF"},   //  VHF
   {  300000000, 3000000000,  0,  7, " UHF"},   //  UHF
   { 3000000000,30000000000,  0,  8, " SHF"}    //  SHF 
};

// Yaesu band voltages = note must be 16 entries to be
// Compatible with the 4 band select lines methods
// This could be better...but not a problem in practice
// The FT817 nominal voltages are: 
    // 160m - 0.33 
    // 80m  - 0.66 
    // 40m  - 1.0  
    // 30m  - 1.3  
    // 20m  - 1.6 
    // 17m  - 2.0 
    // 15m  - 2.3  
    // 12m  - 2.7 
    // 10m  - 3.0 
    // 6m   - 3.3  
    // 2m   - 3.7  
    // 70cm - 4.0  

const float bandVoltageDefs[] = { 
 0.16 ,      //  0 = No band
 0.50 ,      //  1 = 160m
 0.83 ,      //  2 = 80m
 1.15 ,      //  3 = 40m
 1.45 ,      //  4 = 30m
 1.80 ,      //  5 = 20m
 2.15 ,      //  6 = 17m
 2.50 ,      //  7 = 15m
 2.85 ,      //  8 = 12m
 3.15 ,      //  9 = 10m
 3.50 ,      // 10 =  6m
 0.00 ,      // 11 =  4m - dummy for Yaesu
 3.85 ,      // 12 =  2m  
 4.15 ,      // 13 = 70cm
 0.00 ,      // 14 = Dummy
 0.00        // 15 = Dummy   
};

    
